/*
DTWData.java -- DTREND PROJECT
Author: Tero Rontti
Version: 0.1
Revision date: 23.2.2002
Creation date: 10.2.2002


DTREND PROJECT
Project manager: Juha Ylipekkala
Project group: Timo Ahonen, Tero Rontti, Juha Ylipekkala

-------------------------------------------------------------------------------
REVISION HISTORY

Version 0.1
Initial working version
Author: Tero Rontti
Date: 23.2.2002
Known bugs/problems:
*/

package org.dtrend;
import java.util.Vector;

/** A class for handling data storage and DLLApi calls
 * @author Tero Rontti
 * @version 0.1
 */
public class DTWData{
    java.util.Vector Templates = new java.util.Vector();
    DTWDataVector Observation;
    DTWParameters Parameters;
    DTWAllResultsElement[] Results;
    DTWOneResult OneResult;
    int[][] WarpingPath;
    DTWDLLApi DLLApi;

    /** Create a new DTWData object */
    public DTWData( ){
        DLLApi = new DTWDLLApi( );
        Observation = new DTWDataVector();
        Parameters = new DTWParameters();
        Parameters.setDmetrics(Parameters.EuclidianDistanceMetrics);
        Parameters.setSeoffset(0);
        Parameters.setSconstraint(0);
        Parameters.setWwidth(Integer.MAX_VALUE);
    } /** DTWData() ends */

    /** Get templates
     * @return DTWDataVector array containing templates.
     */
    public DTWDataVector[] getTemplates( ){
		int num = Templates.size();
        int template;
		int i;
		
		DTWDataVector[] data = new DTWDataVector[num];
		
		for( i=0; i< num; i++){
            data[i] = new DTWDataVector();
			DTWTemplateListItem item = (DTWTemplateListItem) Templates.get(i);
			data[i].setValues( DLLApi.getTemplate( item.getTIndex() ) );
            data[i].setName( item.getName() );
		}
		return data;
    } /** getVector() ends */
	
    /** Get template
     *  @param tIndexes template indexes: [1] contains DLL index, [0] contains Vector index
     *  @return Template object
     */
    public DTWDataVector getTemplate( int tIndex ){
		DTWTemplateListItem item = (DTWTemplateListItem) Templates.get(tIndex);
		DTWDataVector retval = new DTWDataVector();
		retval.setValues( DLLApi.getTemplate(item.getTIndex()) );
		retval.setName( item.getName() );
		return retval;
    } /** getTemplate() ends */
    
    /** Get template list
     * @return DTWTemplateListItem array containing template list
     */
    public DTWTemplateListItem[] getTemplateList( ){
		int i;
		int num = Templates.size();
		DTWTemplateListItem[] list = new DTWTemplateListItem[num];
		
		for( i=0; i< num; i++){
			list[i] = (DTWTemplateListItem) Templates.get( i );
		}
		return list;
    } /** getTemplateList() ends */
	
    /** Add templates to DLL and Template list
     * @param newTemplates DTWDataVector array to add into data structures
     */
    public void addTemplates( DTWDataVector[] newTemplates ){
		int num = newTemplates.length;
		int i;
		int templateNumber;
		DTWTemplateListItem item;
		
		for( i=0; i< num; i++ ){
			/** Add an entry in template list */
			item = new DTWTemplateListItem( );
			item.setName( newTemplates[i].getName() );
			
			/** Call DLLApi method to add a template and store the template number */
			templateNumber = DLLApi.addTemplate( newTemplates[i].getValues() );
			item.setTIndex( templateNumber );
			Templates.add( item );
		}
    } /** addTemplates() ends */
	
    /** Get name for specified template number
     * @param tIndex Template index
     * @return Template name
     */
    public String getTemplateName( int tIndex ){
		int i;
        return ((DTWTemplateListItem)Templates.get(tIndex)).getName();
    } /** getTemplateName() ends */
	
    /** Delete template
     * @param tIndex Template number to delete
     */
    public void deleteTemplate( int tIndex ){
		int i;
		
		/** Call DLLApi to delete the template */
		DLLApi.delTemplate( ((DTWTemplateListItem)Templates.get(tIndex)).getTIndex() );
        Templates.remove( tIndex );
    } /** deleteTemplate() ends */
	
    /** Delete all templates */
    public void deleteAllTemplates( ){
		int i;
		int num = Templates.size();
		int todel;
		
		for( i=0; i< num; i++){
			/** Delete entry from template list */
			todel = ((DTWTemplateListItem) Templates.get( i )).getTIndex( );
			/** Call DLLApi method to delete template */
			DLLApi.delTemplate( todel );
		}
		
		/** Clear template vector */
		Templates.removeAllElements();
    } /** deleteAllTemplates() ends */
	
    /** Get observation
     * @return Observation DTWDataVector
     */
    public DTWDataVector getObservation( ){
		return Observation;
    } /** getObservation() ends */
	
    /** Set observation
     * @param newObservation New observation
     */
    public void setObservation( DTWDataVector newObservation ){
		Observation = newObservation;
    } /** setObservation() ends */
	
    /** Get observation name
     * @return Observation name in a String
     */
    public String getObservationName( ){
		return Observation.getName();
    } /** getObservationName() ends */
	
    /** Set DTW Parameters
     * @param newParameters New parameters
     */
    public void setDTWParameters( DTWParameters newParameters ){
		Parameters = newParameters;
    } /** setDTWParameters() ends */
	
    /** Get DTW parameters
     * @return Parameters in String format
     */
    public String getDTWParameters( ){
		StringBuffer st = new StringBuffer();
		int se;
		String nl = System.getProperty( "line.separator" );
		
		if( Parameters.getDmetrics() == DTWParameters.EuclidianDistanceMetrics )
			st.append( "Euclidian distance metrics." + nl );
		if( Parameters.getDmetrics() == DTWParameters.DerivativeDistanceMetrics )
			st.append( "Derivative distance metrics." + nl );
		
		st.append( "Warping window width: " + Integer.toString( Parameters.getWwidth() ) + nl );
		
		st.append( "Sconstraint: " + Integer.toString( Parameters.getSconstraint() ) + nl );
		
		se = Parameters.getSeoffset();
		if( se == 0 )
			st.append( "Fixed start and ending points." + nl );
		else
			st.append( "Allowed start and ending point offset: " + Integer.toString( se ) + nl );
		
		return st.toString();
    } /** getDTWParameters() ends */
	
    /** Get Warping window width
     * @return Warping window width parameter
     */
    public int getWindowWidth( ){
		return Parameters.getWwidth();
    } /** getWindowWidth() ends */
	
    /** Compare all templates against the current observation, then compare against the best-matching template */
    public void DTWCompareAll( ){
		Results = new DTWAllResultsElement[Templates.size()];
		int i;
		int j;
		int index;
		int cIndex;
		
        if( Observation.getValues() != null ){
			/** Call DLLApi method to compare all templates */
			Results = DLLApi.CompareAll( Observation.getValues(), Parameters );
			cIndex = Results[0].getTIndex();
			OneResult = DLLApi.CompareOne( Observation.getValues(), Parameters, cIndex );
			for( i=0;i<Results.length;i++){
				for( j=0;j<Templates.size();j++){
					index = Results[i].getTIndex();
					if( index == ((DTWTemplateListItem)Templates.get(j)).getTIndex()){
						Results[i].setTIndex(j);
					}
				}		
			}
			
			OneResult.setTIndex(Results[0].getTIndex());
			/** Call DLLApi method to compare the best matching template against the current observation */
            WarpingPath = OneResult.getWarpingPath();
        }
    } /** DTWCompareAll() ends */
	
    /** Compare one template against the current observation
     * @param tIndex Template number to compare against
     */
    public void DTWCompareOne( int tIndex ){
        DTWAllResultsElement Res = new DTWAllResultsElement();
        OneResult = new DTWOneResult();
        if( Observation.getValues() != null ){
            OneResult = DLLApi.CompareOne( Observation.getValues(), Parameters, ((DTWTemplateListItem)Templates.get(tIndex)).getTIndex());
            WarpingPath = OneResult.getWarpingPath();
            Res.setTIndex(OneResult.getTIndex());
            Res.setDtwDistance(OneResult.getDtwDistance());
            Results = new DTWAllResultsElement[1];
            Results[0] = Res;
        }
    } /** DTWCompareOne() ends */

    /** Get all results
     * @return Results array
     */
    public DTWAllResultsElement[] getResults( ){
		return Results;
    } /** getResults() ends */

    /** Get one result
     * @return OneResult
     */
    public DTWOneResult getOneResult( ){
		return OneResult;
    } /** getOneResult */

    /** Get warping path
     * @return Warping path integer array
     */
    public int[][] getWarpingPath( ){
		return WarpingPath;
    } /** getWarpingPath() ends */

}
