/*
DTWFileIO.java -- DTREND PROJECT
Author: Timo Ahonen
Version: 0.5
Revision date: 15.2.2002
Creation date: 10.3.2002


DTREND PROJECT
Project Manager: Juha Ylipekkala
Project group: Timo Ahonen, Tero Rontti, Juha Ylipekkala

----------------------------------------------------------------------------
REVISION HISTORY:

Version 0.5
-Bugs fixed
-Comments added

Version 0.1 
Initial version
Author: Timo Ahonen
Date: 16.2.2002
Known bugs/problems:
-Too few comments in the code 
*/

package org.dtrend;
import java.io.*;
import java.util.*;
/** This class performs the file input and output operations
 * of the DTREND tool
 * @author Timo Ahonen
 * @version 0.5
 */
public class DTWFileIO {
    private BufferedReader datasetreader;
    private int numlines;
    private String datasetname;
    
    /** Creates new DTWFileIO */
    public DTWFileIO() {
    }

    /** Opens a data set file
     * @param filename Name of the data set file to be opened
     * @throws DTWFileIOException If an I/O exception occurs (e.g. the file does not exist)
     * @return Number of records in the data set file
     */    
    public int openDataSet(String filename) throws DTWFileIOException {
        String readline;

	numlines=0;      
        datasetname=filename; 
        try {
            datasetreader = new BufferedReader(new FileReader(filename));
            while(datasetreader.ready()) { //Count the number of records (lines)
                readline=datasetreader.readLine();
                if (!readline.equals("")) 
		    numlines++;
		else if(datasetreader.ready())
		    throw new DTWFileIOException("Invalid file format (file contains an empty line)");
            }
        } catch(IOException e) {
            System.err.println(e.toString());
            if (e instanceof FileNotFoundException)
            {
                throw(new DTWFileIOException("File "+filename+" not found"));
            } else { //Other type of an error
                throw(new DTWFileIOException("Error opening file "+filename));
            }
        }              
        return numlines;
    }
    
    /** Loads the selected record from the data set file. <CODE>openDataSet</CODE> function must be called prior to calling <CODE>loadRecord</CODE>.
     * @param num Index number of the record to be loaded. Note that the numbering starts from number 1.
     * @throws DTWFileIOException If an exception occurs
     * @return A float array containing the selected record
     */    
    public float[] loadRecord(int num) throws DTWFileIOException {
        StringTokenizer recordst; 
        float[] record;
        Float tempfloat;
        int i;
        
        if(num>numlines)
            throw new DTWFileIOException("Record number "+num+" does not exist");
        try {
            datasetreader.close();
            datasetreader=new BufferedReader(new FileReader(datasetname));
            for(i=1;i<num;i++)
                datasetreader.readLine();
            recordst=new StringTokenizer(datasetreader.readLine());
        } catch (IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error reading data set file");
        }
        try {
            i=0;
            if(recordst.countTokens()<3)
		throw new DTWFileIOException("Record number "+num+" is too short");
	    record=new float[recordst.countTokens()];
            while(recordst.hasMoreTokens())
            {
                record[i]=Float.valueOf(recordst.nextToken()).floatValue();
                i++;
            }
        } catch (NumberFormatException e) {
                System.err.println(e.toString());
                throw new DTWFileIOException("Invalid number format in record "+num);
        }
        return record;
    }
    
    /** Closes the data set file
     * @throws DTWFileIOException If an I/O exception occurs
     */    
    public void closeDataSet() throws DTWFileIOException {
        try {
            datasetreader.close();
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error closing data set file");
        }
    }
    
    /** Loads templates from a text file into a <CODE>DTWData</CODE> object
     * @param filename Name of the template file
     * @param dataobject The object into which the templates are loaded
     * @throws DTWFileIOException If an I/O error occurs or if the file format is invalid
     */    
    public void loadTemplates(String filename, DTWData dataobject) throws DTWFileIOException {
        BufferedReader textreader;
        StringTokenizer linest;
        String readline;
	DTWDataVector thistemplate;
        DTWDataVector[] alltemplates;
        Vector templates=new Vector();
        int i;
        float[] values;
        
        try {
            textreader = new BufferedReader(new FileReader(filename));            
        } catch(IOException e) {
            System.err.println(e.toString());
            if (e instanceof FileNotFoundException)
            {
                throw(new DTWFileIOException("File "+filename+" not found"));
            } else {
                throw(new DTWFileIOException("Error opening file "+filename));
            }
        } //catch
        try {
            while(textreader.ready()) {
                thistemplate = new DTWDataVector();
                readline=textreader.readLine();
		if (!readline.equals("")) {
		    thistemplate.setName(readline);
		    if(textreader.ready()) 
			linest=new StringTokenizer(textreader.readLine());
		    else
			throw new DTWFileIOException("Invalid file format (odd number of lines)");
		    
		    try {
			i=0;
			if(linest.countTokens()<3)
			    throw new DTWFileIOException("Invalid file format (contains a too short template)");
			values=new float[linest.countTokens()];
			while(linest.hasMoreTokens()) {
			    values[i]=Float.valueOf(linest.nextToken()).floatValue();
			    i++;
			} //while
		    } catch (NumberFormatException e) {
			System.err.println(e.toString());
			throw new DTWFileIOException("Invalid number format");
		    
		    } //catch
		    thistemplate.setValues(values);
		    templates.add(thistemplate);
		} else if(textreader.ready())
		    throw new DTWFileIOException("Invalid file format (file contains an empty line)");
            } //while
            textreader.close();
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error reading template file");
        } //catch
        alltemplates=new DTWDataVector[templates.size()];
        templates.toArray(alltemplates);
        dataobject.addTemplates(alltemplates);    
    } //loadTemplates
    
    /** Saves the templates from a <CODE>DTWData</CODE> object into a text file
     * @param filename The name of the file into which the templates are saved
     * @param dataobject The <CODE>DTWData</CODE> object containing the templates
     * @throws DTWFileIOException If an I/O error occurs
     */    
    public void saveTemplates(String filename, DTWData dataobject) throws DTWFileIOException {
        BufferedWriter out;
        DTWDataVector[] alltemplates;
        String outstring;
        float[] values;
        
        try {
            out=new BufferedWriter(new FileWriter(filename,false));
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error opening file "+filename+" for writing");
        } //catch
        
        alltemplates=dataobject.getTemplates();
        
        try {
            for(int i=0;i<alltemplates.length;i++) {
                outstring=alltemplates[i].getName();
                values=alltemplates[i].getValues();
                out.write(outstring,0,outstring.length());
                out.newLine();
                for(int j=0;j<values.length;j++) {
                    outstring=String.valueOf(values[j]);
                    out.write(outstring,0,outstring.length());
                    if(j != values.length-1)
                        out.write(' ');
                } //for
                if(i != alltemplates.length-1)
                    out.newLine();
            } //for

            out.close();
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error writing to file "+filename);
        } //catch
        
    } //saveTemplates
    
    /** Loads an observation from a text file into a <CODE>DTWData</CODE> object
     * @param filename Name of the text file containing the observation
     * @param dataobject The <CODE>DTWData</CODE> object
     * @throws DTWFileIOException If a file I/O error occurs or the file format is invalid
     */    
    public void loadObservation(String filename, DTWData dataobject) throws DTWFileIOException {
        BufferedReader textreader;
        StringTokenizer linest;
        DTWDataVector observation;
        int i;
        float[] values;
        
        try {
            textreader = new BufferedReader(new FileReader(filename));            
        } catch(IOException e) {
            System.err.println(e.toString());
            if (e instanceof FileNotFoundException)
            {
                throw(new DTWFileIOException("File "+filename+" not found"));
            } else {
                throw(new DTWFileIOException("Error opening file "+filename));
            }
        } //catch
        try {
            observation = new DTWDataVector();
            if(textreader.ready()) 
                observation.setName(textreader.readLine());
            else
                throw new DTWFileIOException("Invalid file format");    
            if(textreader.ready()) 
                linest=new StringTokenizer(textreader.readLine());
            else
                throw new DTWFileIOException("Invalid file format");
            
            try {
                i=0;
		if(linest.countTokens()<3)
		    throw new DTWFileIOException("Invalid file format (too short observation)");
                values=new float[linest.countTokens()];
                while(linest.hasMoreTokens()) {
                    values[i]=Float.valueOf(linest.nextToken()).floatValue();
                    i++;
                } //while
            } catch (NumberFormatException e) {
                System.err.println(e.toString());
                throw new DTWFileIOException("Invalid number format");
            } //catch
            observation.setValues(values);
            textreader.close();
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error reading observation file");
        } //catch
        dataobject.setObservation(observation);            
    } //loadObservation
    
    /** Saves the observation from a <CODE>DTWData</CODE> object into a text file
     * @param filename Name of the file into which the observation is saved
     * @param dataobject The <CODE>DTWData</CODE> object containing the observation
     * @throws DTWFileIOException If a file I/O error occurs
     */    
    public void saveObservation(String filename, DTWData dataobject) throws DTWFileIOException {
        BufferedWriter out;
        DTWDataVector observation;
        String outstring;
        float[] values;
        
        try {
            out=new BufferedWriter(new FileWriter(filename,false));
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error opening file "+filename+" for writing");
        } //catch
        
        observation=dataobject.getObservation();
        
        try {
            outstring=observation.getName();
            values=observation.getValues();
            out.write(outstring,0,outstring.length());
            out.newLine();
            for(int j=0;j<values.length;j++) {
                outstring=String.valueOf(values[j]);
                out.write(outstring,0,outstring.length());
                if(j != values.length-1)
                    out.write(' ');
            } //for
            out.close();
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error writing to file "+filename);
        } //catch          
    } //saveObservation
    
    /** Saves the DTW algorithm run information from a <CODE>DTWData</CODE> object into a log file
     * @param filename Name of the log file
     * @param dataobject The <CODE>DTWData</CODE> object containing the DTW information
     * @throws DTWFileIOException If a file I/O error occurs
     */    
    public void saveLog(String filename, DTWData dataobject) throws DTWFileIOException {
        BufferedWriter out;
        String outstring;
        Date timestamp=new Date();
        DTWAllResultsElement[] results;
        int[][] warpingpath;
        
        try {
            out=new BufferedWriter(new FileWriter(filename,false));
        } catch(IOException e) {
            System.err.println(e.toString());
            throw new DTWFileIOException("Error opening file "+filename+" for writing");
        } //catch
                        
        try {
            outstring="DTREND LOG FILE";
            out.write(outstring,0,outstring.length());
            out.newLine();
            
            outstring=timestamp.toString();
            out.write(outstring,0,outstring.length());
            out.newLine();
            out.newLine();
            
            outstring="DTW Algorithm Parameters:";
            out.write(outstring,0,outstring.length());
            out.newLine();
            
            outstring=dataobject.getDTWParameters(); //getDTWParameters returns the parameters in such a format that they can be written directly to the log file
            out.write(outstring,0,outstring.length());
            out.newLine();
            out.newLine();
            
            outstring="Observation: "+dataobject.getObservationName();
            out.write(outstring,0,outstring.length());
            out.newLine();
            out.newLine();
            
            outstring="Results:";
            out.write(outstring,0,outstring.length());
            out.newLine();
            
            results=dataobject.getResults();
            for(int i=0;i<results.length;i++) { //Prints the results into the log file
                outstring=dataobject.getTemplateName(results[i].getTIndex()); //Name of the template
                outstring+=": "+String.valueOf(results[i].getDtwDistance());  //and the dtw distance 
                out.write(outstring,0,outstring.length());
                out.newLine();
            } //for
            out.newLine();
            
            outstring="Warping path (template "+dataobject.getTemplateName(results[0].getTIndex())+"):";
            out.write(outstring,0,outstring.length());
            out.newLine();
            
            warpingpath=dataobject.getWarpingPath();
            for(int i=0;i<warpingpath.length;i++) { //Prints the warping path into the log file
                outstring="("+warpingpath[i][0]+","+warpingpath[i][1]+")";
                out.write(outstring,0,outstring.length());
                out.newLine();
            } //for
            
            out.close();
        } catch(IOException e) {
             System.err.println(e.toString());
            throw new DTWFileIOException("Error writing to file "+filename);
        } //catch                     
    } //saveLog
} //DTWFileIO
